# Mustache Logic-less Templates

Integrates Mustache.php plus Mustache.js into Drupal.

Includes a generic framework to build DOM content, which can be automatically
synchronized on the client-side. Any RESTful Json endpoint can be used for
automatic content synchronization.

Also includes a text filter plugin that enables users to use Mustache template
syntax for creating dynamic content. More on that see section "About the
Mustache text filter".

Using Mustache.php by
 * Copyright (c) 2010-2015 Justin Hileman

Using Mustache.js by
* Copyright (c) 2009 Chris Wanstrath (Ruby)
* Copyright (c) 2010-2014 Jan Lehnardt (JavaScript)
* Copyright (c) 2010-2015 The mustache.js community

Using morphdom by
* Copyright (c) Patrick Steele-Idem <pnidem@gmail.com> (psteeleidem.com)

More about Mustache: http://mustache.github.io/

## Requirements

The following libraries are required:
- Mustache.php by bobthecow v2.13.0 or newer as vendor library
  https://github.com/bobthecow/mustache.php
  The vendor library should be installed automatically if you have a Composer
  based workflow. If not, you need to install it manually. It is generally
  recommended to use Composer to install this library:
  `composer require "mustache/mustache:^2.13"` or have a look at
  https://www.drupal.org/docs/develop/using-composer/using-composer-to-install-drupal-and-manage-dependencies
- Mustache.js by janl v4.2.0 or newer as Drupal Javascript library
  https://github.com/janl/mustache.js
  A minified version of this library is already included in this module, so
  it is not required to download it manually. It might be that this module
  does not include the most recent version of the Javascript library though.
  You can always download the library on your own and place it into the Drupal
  app root directory <DRUPAL_APP_ROOT>/libraries/js/mustache/mustache.min.js.
  If you want to download a minified version, it is available in the npm
  repository https://www.npmjs.com/package/mustache.

## Recommendations

It's recommended to use a Composer-based workflow to install
any dependency of this module.

## How Mustache templating works in Drupal

The usage of Mustache templates is mainly defined via render arrays.
Therefore, the module provides a special render element called 'mustache'.

Mustache templates can be put besides Twig templates
into the templates directory of your theme. A file which is a Mustache template
must end with .mustache.tpl.

Example: A template for rendering social share buttons could be named
as "social-share-buttons.mustache.tpl" and put into yourtheme/templates/.
To render the Mustache template, a render array can be created as follows:

```php
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $data,
];
```

Note that `#template` does not contain the .mustache.tpl file ending.
See also section "Render array examples" below to find out
how to create render arrays for Mustache templates.

In case a module wants to define a Mustache template, it can do so by
implementing `hook_mustache_templates()`. With this hook, you can also
define default values for building the render array. All Mustache template hooks
are documented in the mustache.api.php file of this module.

You don't enjoy writing render arrays? Try out and have a look at the section
"Using the helper class for rendering Mustache templates" below.

## Render array examples

You can either build render arrays directly, or build them with the helper
class `MustacheRenderTemplate`.

Using the helper class is preferable, because it's covered by an automated
unit test, and since it's providing a more stable and convenient API,
it will support you on automatically covering upcoming changes
regards building the render array.

```php
// Example 1:
// Renders the template (only) via Mustache.php.
// #data can either be an array or a url pointing to the Json data source.
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $data,
];
// Equivalent using the helper class:
$render = $render = MustacheRenderTemplate::build('social_share_buttons')
->usingData($data)
->toRenderArray();

// Example 2:
// With #select you can pass a certain subset of nested data to the template.
$nested = ['foo' => ['bar' => ['baz' => 'qux']]];
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $nested,
  '#select' => ['foo', 'bar'], // Would pass ['baz' => 'qux']
];

// Example 3:
// Renders the template via Mustache.php,
// and refreshes the content via Mustache.js one time per second.
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $url,
  '#sync' => [
    'period' => 1000, // Milliseconds.
    'limit' => -1, // Unlimited number of refreshes.
  ],
];

// Example 4:
// Renders a placeholder, which will be replaced by the given
// template via Mustache.js using Json data supplied by $url.
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $url,
  '#placeholder' => ['#markup' => '<span>Loading...</span>'],
  '#sync' => [
    'wrapper_tag' => 'span', // Set wrapper tag to be a span (default is div).
  ],
];

// Example 5:
// Renders the template via Mustache.php using $data, and synchronizes
// the content one time via Mustache.js using Json data supplied by $url.
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $data,
  '#sync' => [
    'data' => $url,
  ],
];

// Example 6:
// Just like example 5, but this time the template would only be synchronized,
// when a click event has been triggered on any element with a .button CSS class.
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $data,
  '#sync' => [
    'data' => $url,
    'trigger' => [['.button', 'click', 3]], // Would run up to 3 times. -1 would be unlimited.
  ],
];
// Equivalent using the helper class:
$render = MustacheRenderTemplate::build('social_share_buttons')
->usingData($data)
->withClientSynchronization()
  ->usingDataFromUrl($url)
  ->startsWhenElementWasTriggered('.button')
    ->atEvent('click')
    ->upToNTimes(3)
->toRenderArray();
```

## Insertion of synchronized content

By default, the synchronization replaces previously generated content.
You can change this behavior by setting the `adjacent` option, which defines
how rendered content should be inserted:

```php
$render = [
  '#type' => 'mustache',
  '#template' => 'my_item_list',
  '#data' => $data,
  '#sync' => [
    'adjacent' => 'beforeend',
  ],
];
// Equivalent using the helper class:
$render = MustacheRenderTemplate::build('my_item_list')
->usingData($data)
->withClientSynchronization()
  ->usingDataFromUrl($url)
  ->insertsAt('beforeend')
->toRenderArray();
```

The position can be either `beforebegin`, `afterbegin`, `beforeend` or
`afterend`. Have a look at the Javascript function `Element.insertAdjacentHTML`
for more information about inserting HTML with the position parameter.

## Script execution

By default - due to possible security implications - scripts inside synchronized
DOM content will not be executed automatically. However, if you want to
let the synchronization process also execute possible included scripts,
you can do so by setting the `eval` parameter to TRUE:

```php
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $data,
  '#sync' => [
    'data' => $url,
    'eval' => TRUE,
  ],
];
// Equivalent using the helper class:
$render = MustacheRenderTemplate::build('social_share_buttons')
->usingData($data)
->withClientSynchronization()
  ->usingDataFromUrl($url)
  ->executesInnerScripts(TRUE)
->toRenderArray();
```

## Morphing DOM trees

When using client-side synchronization, by default the DOM content is being
completely replaced. While replacing an existing DOM tree with an entirely new
DOM tree is fast, all of the internal state associated with the existing DOM
nodes will be lost. Instead of always replacing the existing DOM tree with a new
DOM tree, you can use the morphdom framework for DOM tree transformations:

```php
$render = [
  '#type' => 'mustache',
  '#template' => 'social_share_buttons',
  '#data' => $data,
  '#sync' => [
    'data' => $url,
    'morph' => TRUE,
  ],
];
```

Please note that you cannot use the `morph` setting together with the
`adjacent` setting. When `adjacent` is set, the `morph` setting will be ignored.

## Auto incrementing

The synchronization framework provides the ability to auto increment on
a certain url query param or data key:

```php
$render = MustacheRenderTemplate::build('my_paged_list')
->withClientSynchronization()
  ->usingDataFromUrl($url)
  ->increments()
->toRenderArray();
```

The `increments()` method enables you to further define the auto increment,
including the step size to increment, on which query param to run incrementing,
how many times it should increment, and whether it should loop or not.

## Form binding

Form binding enables you to build simple reactive forms without writing any
Javascript.

You can bind a form to the Mustache render build, so that the form values are
directly provided within the `form` key of the data that is being passed to the
template rendering. For example, if you have an input field of name `search`,
then you can access that value in the template with `{{form.search}}`. When a
url is set for fetching Json-encoded data, then the form values will be passed
to that endpoint as query string parameters. This enables your Json endpoint to
return any requested data for the given form input.

The following example binds a form for client synchronization:

```php
$render = [];
$form = <<<HTML
<div>
    <form id="myform" method="GET" action="">
      <input id="myfield" type="text" name="search" value="" />
    </form>
</div>
HTML;
$build = MustacheRenderTemplate::build('inline_result', '<b>Submitted</b>: {{form.search}}<br /><b>Received</b>: {{echo.search}}');
$build->withPlaceholder(['#markup' => '']); // This is an empty placeholder to skip server-side rendering.
$build->usingDataFromUrl('/my-echo-feed'); // Form values will be automatically passed as query string parameters.
$build->withClientSynchronization()
  ->morphing() // Morphing DOM trees also work with form binding.
  ->usingFormValues('#myform')
  ->startsWhenElementWasTriggered('#myfield')
  ->atEvent('input')
  ->always();

$render[] = $build->toRenderArray();
$render[] = ['#markup' => \Drupal\mustache\Render\Markup::create($form)];
```

The above example skips server-side rendering, but form binding also works with
Mustache.php rendering. To include the form binding also for Mustache.php,
use `\Drupal\mustache\Helpers\MustacheRenderTemplate::usingFormValues`, like the
following example:

```php
$build = MustacheRenderTemplate::build('inline_result', '<b>Submitted</b>: {{form.search}}<br /><b>Received</b>: {{echo.search}}');
$build->usingDataFromUrl('/mustache-test/echo-feed')
  ->usingFormValues($form_state)
  ->withClientSynchronization()
    ->morphing()
    ->startsWhenElementWasTriggered('#myfield')
    ->atEvent('input')
    ->always();
  ;
```

As seen in the example above, you can directly pass a form state object (an
instance of `Drupal\Core\Form\FormStateInterface`). The render build takes care
for automatically binding the form and passes current form values to the
Json endpoint.

It should be noted, when working with Drupal forms, mostly the form will have a
redirect after form submission. When a redirect happens, then the form values
are no longer available. The form redirect may be either disabled, or the
relevant arguments may be attached to the redirect url, then extracted and
(manually) passed to the data array.

For more examples on how to bind forms, have a look at the class
`\Drupal\mustache_test\Form\MustacheTestForm` within the mustache_test module
folder, and have a look at the documented method
`\Drupal\mustache\Helpers\MustacheRenderTemplate::usingFormValues`.

## Javascript events

The following events are provided at the specified synchronization targets:

- `mustacheSyncBegin`: Triggered when the synchronization for
  a specific item has just been started.
- `mustacheSyncFinish`: Triggered when the synchronization for
  a specific item has just been finished.

## Inline templates

You can write templates on-the-fly by directly defining the template content:
```php
$render = MustacheRenderTemplate::build('my_inline_template', '{{foo}}')
->usingData(['foo' => 'bar'])
->toRenderArray();
```

This might be useful if you want to quickly try out Mustache templates, or you
need to dynamically create template contents. It should be noted that this
method involves a different caching strategy that is not that efficient as
when using registered template files. That is why it is generally recommended
to always use registered template files when possible.

## Token support

You can enable Tokens to be included like Mustache variables, and provide custom
data and options for the Token replacement processor:
```php
$render = MustacheRenderTemplate::build('node_inline', '{{node.title}}')
->withTokens(['node' => $node])
->toRenderArray();
```

By default, this processing clears empty token values. You can change that
behavior by explicitly setting the options for the Token replacement processor.
The processing also takes care for loading entities from the global context,
so you do not need to set those values manually. But you can still pass through
your own data.

## About the Mustache text filter

For trusted users with advanced permissions (e.g. technical admins or trusted
editors), a text filter plugin is available, that enables users to write
Mustache template syntax for creating dynamic content. This text processor
includes token support out of the box, thus not requiring an extra module to
also include tokens. If you also want to use that plugin on other entities than
nodes, then the contrib Token module (drupal.org/project/token) should be
additionally installed.

The Mustache filter enables you to use tokens in a more dynamic way, e.g.
enclosing them with a condition:

```html
<h1>Welcome to {{site.name}}</h1>
{{#site.slogan}}
  <h2>{{site.slogan}}</h2>
{{/site.slogan}}
```

Beware that this templating engine enables you to bypass escaping by using
the triple bracket syntax `{{{}}}`. Because of that you should only give trusted
users the permission to use that filter.

It is very likely that this filter plugin does not work well in combination
with other filters, that involve any further dynamic processing like another
token filter plugin. For example, the token filter by the Token Filter module
should not be used together with the Mustache filter on the same text format.

## Internal module development notices

The minified sync.js was generated with the help of terser
(https://www.npmjs.com/package/terser) using the command
`terser sync.js -o sync.min.js -c -m`

The minified version of the Mustache.js library including its license file was
obtained from the npm repository https://www.npmjs.com/package/mustache.

The minified version of the morphdom library including its license file was
obtained from the npm repository https://www.npmjs.com/package/morphdom.
