<?php

namespace Drupal\mustache\Element;

use Drupal\Component\Utility\Crypt;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\BubbleableMetadata;
use Drupal\Core\Render\Element\RenderElement;
use Drupal\Core\Template\Attribute;
use Drupal\Core\Url;
use Drupal\mustache\Exception\MustacheException;
use Drupal\mustache\MustachePhpLoader;
use Drupal\mustache\Render\Markup;
use GuzzleHttp\Exception\GuzzleException;

/**
 * Provides an element for rendering Mustache templates.
 *
 * @RenderElement("mustache")
 */
class Mustache extends RenderElement {

  /**
   * Statically cached context IDs.
   *
   * @var string[]
   */
  protected static $contextIds;

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = get_class($this);
    return [
      '#theme' => 'mustache',
      '#pre_render' => [
        [$class, 'generateContentMarkup'],
      ],
    ];
  }

  /**
   * Generates content markup for a Mustache template.
   *
   * @param array $element
   *   The element to render.
   *
   * @return array
   *   The element to render.
   *
   * @throws \Exception
   *   In case something went wrong on placeholder rendering,
   *   or when the template or data url could not be found.
   */
  public static function generateContentMarkup(array $element) {
    if (!isset($element['#cache'])) {
      // Caching will be bubbled up if necessary.
      $element['#cache'] = [];
    }

    /** @var \Drupal\mustache\MustacheTemplates $templates */
    $templates = \Drupal::service('mustache.templates');
    /** @var \Drupal\mustache\Summable\SummableScriptsInterface $summables */
    $summables = \Drupal::service('mustache.summables');

    // Load and encode the template content.
    $template_name = $element['#template'];
    if (isset($element['#inline']) && ($element['#inline'] !== FALSE)) {
      $template_content = $element['#inline'];
      $element_defaults = [];
      $element['#summable'] = FALSE;
      $use_inline_cache = TRUE;
    }
    else {
      $template_content = $templates->getContent($template_name);
      $element_defaults = $templates->getElementDefaults($template_name);
      $element['#inline'] = FALSE;
      $use_inline_cache = FALSE;
    }

    // Set default element values.
    $sync_defaults = isset($element_defaults['#sync']) ? $element_defaults['#sync'] : [];
    unset($element_defaults['#sync']);
    $element += $element_defaults;

    static::processTokens($element, $template_content);
    $template_encoded = trim(substr(json_encode($template_content, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE | JSON_HEX_TAG | JSON_HEX_APOS | JSON_HEX_AMP | JSON_HEX_QUOT), 1, -1));

    // Attach any retrieved information about the template.
    $element['#template'] = [
      'name' => $template_name,
      'content' => $template_content,
      'encoded' => $template_encoded,
      'summable' => isset($element['#summable']) ? (bool) $element['#summable'] : $summables->isEnabled(),
    ];

    if (!isset($element['#partials'])) {
      $element['#partials'] = [];
    }
    if (!isset($element['#inline_partials'])) {
      $element['#inline_partials'] = [];
    }

    // Begin to determine the data to handle with.
    $data = isset($element['#data']) ? $element['#data'] : [];
    $select = isset($element['#select']) ? $element['#select'] : NULL;
    if (is_string($select)) {
      $select = [$select];
    }

    // Obtain the Json data url, if given.
    if ($url = static::getUrlFromParam($data)) {
      $data = NULL;
    }

    // Give a last chance for still being able to use data.
    if (isset($element['#override_data'])) {
      $data = $element['#override_data'];
    }

    // Prepare form values, if given.
    if (!isset($element['#form'])) {
      $element['#form'] = [];
    }
    if (!empty($element['#form'])) {
      if ($element['#form'] instanceof FormStateInterface) {
        $form_state = $element['#form'];
        $element['#form'] = [];
        $form_array = &$form_state->getCompleteForm();
        if (!isset($form_array['#id'])) {
          $form_id = $form_state->getFormObject() ? $form_state->getFormObject()->getFormId() : Crypt::randomBytesBase64(4);
          $form_array['#id'] = Html::getUniqueId($form_id);
          $form_array['#attributes']['data-drupal-selector'] = Html::getId($form_id);
        }
        if (isset($form_array['#attributes']['data-drupal-selector'])) {
          $element['#form']['selector'] = '[data-drupal-selector="' . $form_array['#attributes']['data-drupal-selector'] . '"]';
        }
        $element['#form']['values'] = $form_state->getValues();
        unset($element['#form']['values']['form_build_id']);
      }
      elseif (!is_array($element['#form'])) {
        if (is_string($element['#form'])) {
          $element['#form'] = ['selector' => $element['#form']];
        }
        else {
          // A last attempt to get an array. If it fails, it should fail hard.
          $element['#form'] = (array) $element['#form'];
        }
      }
      if (isset($url) && !empty($element['#form']['values'])) {
        $url = static::rebuildUrl($url, $element['#form']['values']);
      }
      elseif (!isset($element['#form']['values'])) {
        $element['#form']['values'] = [];
      }
    }

    if (!empty($element['#sync'])) {
      $element['#use_sync'] = TRUE;
      $use_morph = FALSE;
      $sync_options = &$element['#sync'];
      $sync_options += $sync_defaults;

      if (!isset($sync_options['items'])) {
        $extract_keys = [
          'eval',
          'behaviors',
          'data',
          'select',
          'increment',
          'delay',
          'period',
          'limit',
          'trigger',
          'adjacent',
          'morph',
          'form',
        ];
        $sync_item = [];
        foreach ($extract_keys as $extract_key) {
          if (isset($sync_options[$extract_key])) {
            $sync_item[$extract_key] = $sync_options[$extract_key];
            unset($sync_options[$extract_key]);
          }
        }
        $sync_options['items'] = [$sync_item];
      }

      foreach ($sync_options['items'] as &$sync_item) {
        // Check for the Json data provider url, if given.
        $sync_url = NULL;
        if (!empty($sync_item['data'])) {
          $sync_url = static::getUrlFromParam($sync_item['data']);
        }
        elseif (isset($url)) {
          $sync_url = clone $url;
        }
        else {
          $sync_item['data'] = $data ?: [];
        }

        // Include nested selection, if specified.
        $sync_select = $select;
        if (!empty($sync_item['select'])) {
          $sync_select = $sync_item['select'];
          if (is_string($sync_select)) {
            $sync_select = [$sync_select];
          }
        }
        if (!empty($sync_select)) {
          $sync_item['select'] = json_encode($sync_select, JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE | JSON_HEX_TAG | JSON_HEX_APOS | JSON_HEX_AMP | JSON_HEX_QUOT);
        }

        // Include auto increment, if specified.
        if (isset($sync_item['increment']) && $sync_item['increment'] !== FALSE) {
          if (!is_array($sync_item['increment'])) {
            $sync_item['increment'] = [];
          }
          $sync_item['increment'] = json_encode($sync_item['increment']);
        }

        // Include (non-)execution of Drupal behaviors.
        if (isset($sync_item['behaviors'])) {
          $sync_item['behaviors'] = empty($sync_item['behaviors']) ? 'false' : 'true';
        }

        // Include and encode triggering elements, if specified.
        if (isset($sync_item['trigger'])) {
          foreach ($sync_item['trigger'] as &$trigger) {
            $trigger[1] = !empty($trigger[1]) && is_string($trigger[1]) ? $trigger[1] : 'load';
            $trigger[2] = !empty($trigger[2]) && is_int($trigger[2]) ? $trigger[2] : 1;
            if ($trigger[2] < 0) {
              // Always use -1 for triggering without any limits.
              $trigger[2] = -1;
            }
          }
        }
        if (!empty($sync_item['trigger'])) {
          $sync_item['trigger'] = json_encode($sync_item['trigger']);
        }

        // Validate and encode the adjacent option, if specified.
        if (!isset($sync_item['adjacent'])) {
          $sync_options['adjacent'] = NULL;
          $sync_item['adjacent'] = NULL;
        }
        elseif (!in_array($sync_item['adjacent'],
          ['beforebegin', 'afterbegin', 'beforeend', 'afterend'])) {
          throw new MustacheException(t('Invalid adjacent position :option given. See the Javascript documentation about Element.insertAdjacentHTML which position options are available.', [':option' => $sync_item['adjacent']]));
        }
        elseif (!isset($sync_options['adjacent'])) {
          $sync_options['adjacent'] = $sync_item['adjacent'];
        }
        if (!empty($sync_item['adjacent'])) {
          $sync_item['adjacent'] = json_encode($sync_item['adjacent']);
          unset($sync_item['morph']);
        }

        if (!empty($sync_item['morph'])) {
          $sync_item['morph'] = json_encode($sync_item['morph']);
          $use_morph = TRUE;
        }

        if (!isset($sync_item['form']) && !empty($element['#form'])) {
          if (isset($element['#form']['selector']) && empty($element['#form']['values'])) {
            $sync_item['form'] = $element['#form']['selector'];
          }
          elseif (isset($element['#form']['values'])) {
            $sync_item['form'] = [
              'selector' => isset($element['#form']['selector']) ? $element['#form']['selector'] : NULL,
              'el' => NULL,
              'values' => &$element['#form']['values'],
            ];
          }
        }
        if (!empty($sync_item['form'])) {
          if (isset($sync_url) && !empty($sync_item['form']['values'])) {
            $sync_url = static::rebuildUrl($sync_url, $sync_item['form']['values']);
          }
          $sync_item['form'] = json_encode($sync_item['form'], JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE | JSON_HEX_TAG | JSON_HEX_APOS | JSON_HEX_AMP | JSON_HEX_QUOT | JSON_FORCE_OBJECT);
        }
        if (isset($sync_item['max_age'])) {
          $sync_item['max_age'] = (int) $sync_item['max_age'];
          if ($sync_item['max_age'] < 0) {
            // Negative values do not make sense for max_age.
            unset($sync_item['max_age']);
          }
        }

        if (isset($sync_url)) {
          $sync_url->setAbsolute($sync_url->isExternal());
          $sync_item['url'] = $sync_url;
          $sync_item['data'] = $sync_url->toString();
        }
        $sync_item['data'] = json_encode($sync_item['data'], JSON_UNESCAPED_SLASHES | JSON_UNESCAPED_UNICODE | JSON_HEX_TAG | JSON_HEX_APOS | JSON_HEX_AMP | JSON_HEX_QUOT | JSON_FORCE_OBJECT);
      }

      // Build and attach the structure for DOM content synchronization.
      $element_id = 'msync-' . Crypt::randomBytesBase64(4);
      $attributes = new Attribute(
        ['id' => $element_id, 'class' => ['mustache-sync', 'not-synced']]);
      $sync_options['attributes'] = $attributes;
      $sync_options['wrapper_tag'] = !empty($sync_options['wrapper_tag']) ? $sync_options['wrapper_tag'] : 'div';
      if ($use_morph) {
        $element['#attached']['library'][] = 'mustache/morphdom';
      }
      if ($element['#template']['summable']) {
        $element['#attached']['library'][] = $summables->getLibraryName($template_name);
      }
      else {
        $element['#attached']['library'][] = 'mustache/sync';
      }

      // Add further, custom defined attributes.
      if (isset($element['#attributes'])) {
        if (isset($element['#attributes']['class'])) {
          $attributes->addClass($element['#attributes']['class']);
          unset($element['#attributes']['class']);
        }
        foreach ($element['#attributes'] as $attr_name => $attr_value) {
          $attributes->setAttribute($attr_name, $attr_value);
        }
        unset($element['#attributes']);
      }

      if (!empty($element['#partials'])) {
        foreach ($element['#partials'] as $name) {
          $partial_defaults = $templates->getElementDefaults($name);
          $summable = isset($partial_defaults['#summable']) ? (bool) $partial_defaults['#summable'] : $summables->isEnabled();
          if ($summable && (empty($element_defaults['#partials']) || !in_array($name, $element_defaults['#partials']))) {
            $library_name = $summables->getLibraryName($name);
            if (!in_array($library_name, $element['#attached']['library'])) {
              array_unshift($element['#attached']['library'], $library_name);
            }
          }
          elseif (!$summable && !isset($element['#inline_partials'][$name])) {
            $element['#inline_partials'][$name] = [
              '#type' => 'mustache_template_js_inline',
              '#template' => $name,
            ];
          }
        }
      }
    }
    else {
      $element['#use_sync'] = FALSE;
      $element['#sync'] = [];
    }

    if (!empty($element['#placeholder'])) {
      // Instead of rendering the content via Mustache.php,
      // render an arbitrary placeholder, by respecting its cache metadata.
      $element_metadata = CacheableMetadata::createFromRenderArray($element);
      $placeholder_metadata = CacheableMetadata::createFromRenderArray($element['#placeholder']);
      $element_metadata->merge($placeholder_metadata)->applyTo($element);

      /** @var \Drupal\Core\Render\Renderer $renderer */
      $renderer = \Drupal::service('renderer');
      $rendered = $renderer->render($element['#placeholder']);
      $element['#content'] = Markup::create($rendered);
    }
    else {
      // Render the content via Mustache.php.
      if (!isset($data) && isset($url)) {
        // Rendering via Mustache.php requires some data,
        // which now must be fetched from the url.
        $client = \Drupal::httpClient();
        $uri = $url->setAbsolute(TRUE)->toString();
        try {
          $response = $client->request('GET', $uri);
        }
        catch (GuzzleException $e) {
          \Drupal::logger('mustache')->error(t('Failed to fetch data from @url. Exception message: @message',
            ['@url' => $uri, '@message' => $e->getMessage()]));
          return static::emptyElement();
        }

        if (!($response->getStatusCode() == 200)) {
          \Drupal::logger('mustache')->error(t('Failed to fetch data from @url. Returned status code was: @status',
            ['@url' => $uri, '@status' => $response->getStatusCode()]));
          return static::emptyElement();
        }

        $body = $response->getBody()->getContents();
        $data = json_decode($body, TRUE);
        if (!isset($data)) {
          \Drupal::logger('mustache')->error(t('Invalid Json received from from @url.', ['@url' => $uri]));
          return static::emptyElement();
        }
      }

      if (!empty($select) && !empty($data)) {
        if (!is_array($data)) {
          // In case something different was passed,
          // try to convert it. When it fails, it should fail hard.
          $data = (array) $data;
        }
        $subset_exists = NULL;
        $data = NestedArray::getValue($data, $select, $subset_exists);
        if (!$subset_exists) {
          // Subset must be an array.
          $data = [];
        }
      }

      if (isset($element['#form']['values'])) {
        $data['form'] = &$element['#form']['values'];
      }

      /** @var \Drupal\mustache\MustachePhpEngine $mustache */
      $mustache = \Drupal::service('mustache.engine');
      $partials_loader = $mustache->getPartialsLoader();

      if (!empty($element['#inline_partials']) && ($partials_loader instanceof \Mustache_Loader_MutableLoader)) {
        foreach ($element['#inline_partials'] as $name => $value) {
          if (is_string($value)) {
            $use_inline_cache = TRUE;
            $partials_loader->setTemplate($name, $value);
          }
          elseif (is_array($value)) {
            if (isset($value['#template'])) {
              $name = $value['#template'];
            }
            if (isset($value['#inline']) && $value['#inline'] !== FALSE) {
              $use_inline_cache = TRUE;
              $partials_loader->setTemplate($name, $value['#inline']);
            }
            elseif (!in_array($name, $element['#partials'])) {
              $element['#partials'][] = $name;
            }
          }
        }
      }

      if ($use_inline_cache) {
        $template = $element['#inline'] === FALSE ? $mustache->loadTemplateByName($template_name) : NULL;
        foreach ($element['#partials'] as $name) {
          $mustache->loadTemplateByName($name);
        }
        $engine_cache = $mustache->getCache();
        /** @var \Drupal\mustache\MustachePhpInlineCache $inline_cache */
        $inline_cache = \Drupal::service('mustache.inline_php_cache');
        $inline_cache->setEngineCache($engine_cache);
        $mustache->setCache($inline_cache);
        if (isset($template)) {
          $output = $template->render($data);
        }
        else {
          $loader = $mustache->getLoader();
          if ($loader instanceof \Mustache_Loader_MutableLoader) {
            $loader->setTemplate($template_name, $template_content);
            $output = $mustache->loadTemplateByName($template_name)->render($data);
          }
          else {
            $output = $mustache->loadTemplate($template_content)->render($data);
          }
        }
        $mustache->setCache($engine_cache);
      }
      else {
        $output = $mustache->loadTemplateByName($template_name)->render($data);
      }
      $element['#content'] = Markup::create($output);

      if ($element['#use_sync'] && ($partials_loader instanceof MustachePhpLoader)) {
        // In general, partials should be defined within 'default' values via
        // hook_mustache_templates(). This section is a last resort to extract
        // used partials during server-side rendering, which might not be known
        // from the info array yet. It should be noted that partials can be
        // loaded at runtime, and as a consequence this section is not
        // considered to include every possible partial, as it might be that a
        // partial is not included (e.g. when a certain condition was not met
        // for including a certain partial).
        foreach ($partials_loader->getLoadedRegistryTemplateNames() as $name) {
          if (($name === $template_name) || in_array($name, $element['#partials'])) {
            continue;
          }
          $partial_defaults = $templates->getElementDefaults($name);
          $summable = isset($partial_defaults['#summable']) ? (bool) $partial_defaults['#summable'] : $summables->isEnabled();
          if ($summable) {
            $library_name = $summables->getLibraryName($name);
            if (!in_array($library_name, $element['#attached']['library'])) {
              array_unshift($element['#attached']['library'], $library_name);
            }
          }
          elseif (!isset($element['#inline_partials'][$name])) {
            $element['#inline_partials'][$name] = [
              '#type' => 'mustache_template_js_inline',
              '#template' => $name,
            ];
          }
        }
        foreach ($partials_loader->getLoadedNotRegistryTemplateNames() as $name) {
          // On-the fly generated templates will only be loaded inline,
          // because it is not known if they may vary.
          if (!isset($element['#inline_partials'][$name])) {
            $element['#inline_partials'][$name] = [
              '#type' => 'mustache_template_js_inline',
              '#template' => $name,
              '#inline' => $partials_loader->load($name),
            ];
          }
        }
        $partials_loader->resetLoadedRegistryTemplateList();
        $partials_loader->resetLoadedNotRegistryTemplateList();
      }
    }

    foreach ($element['#inline_partials'] as $name => $value) {
      if (($name === $template_name) && (!$element['#summable'] || $element['#inline'] !== FALSE)) {
        // The scoped template will already be added as inline Javascript,
        // thus no need to include it also as partial.
        unset($element['#inline_partials'][$name]);
        continue;
      }
      // Server-side rendering is already done at this point, which means
      // that inline templates would be now used at client-side only (if
      // enabled). For that, the template content is being wrapped as inline
      // Javascript from here.
      if (is_string($value)) {
        $element['#inline_partials'][$name] = [
          '#type' => 'mustache_template_js_inline',
          '#template' => $name,
          '#inline' => $value,
        ];
      }
    }

    return $element;
  }

  /**
   * Returns an empty element which is not cacheable.
   *
   * This element would only be used at unusual circumstances.
   * Because it's not what the result should look like,
   * the caching is completely disabled for being able
   * to generate an expected result as soon as possible.
   *
   * @return array
   *   The empty element.
   */
  public static function emptyElement() {
    return [
      '#cache' => ['max-age' => 0],
      '#printed' => TRUE,
      '#markup' => Markup::create(''),
    ];
  }

  /**
   * Helper function to get the url from the given parameter.
   *
   * @param mixed $param
   *   The given parameter.
   *
   * @return \Drupal\Core\Url|null
   *   The url object, if found.
   */
  protected static function getUrlFromParam($param) {
    if ($param instanceof Url) {
      return $param;
    }
    elseif (is_string($param)) {
      try {
        return Url::fromUri($param);
      }
      catch (\InvalidArgumentException $e) {
        return Url::fromUserInput($param);
      }
    }
    return NULL;
  }

  /**
   * Processes tokens, if any, for the given template content.
   *
   * @param array &$element
   *   The current element build array.
   * @param string &$template_content
   *   The current template content.
   */
  protected static function processTokens(array &$element, &$template_content) {
    if (!isset($element['#with_tokens']) || $element['#with_tokens'] === FALSE) {
      // Tokens are not enabled for this build, so abort.
      return;
    }

    if (!isset($element['#data'])) {
      $element['#data'] = [];
    }
    if (!is_array($element['#with_tokens'])) {
      $element['#with_tokens'] = [];
    }

    $token_service = \Drupal::token();
    $token_settings = &$element['#with_tokens'];
    $token_options = isset($token_settings['options']) ? $token_settings['options'] : [];

    // Build up the token data.
    $token_data = isset($token_settings['data']) ? $token_settings['data'] : [];
    /** @var \Drupal\token\TokenEntityMapperInterface $entity_token_type_mapper */
    $entity_token_type_mapper = \Drupal::hasService('token.entity_mapper') ? \Drupal::service('token.entity_mapper') : NULL;
    // Merge with static data, e.g. provided by preprocess hooks.
    if ($static_token_data = &drupal_static('mustache_tokens', [])) {
      if (isset($entity_token_type_mapper)) {
        foreach ($static_token_data as $type => $value) {
          if (!($value instanceof EntityInterface)) {
            continue;
          }
          $the_real_token_type = $entity_token_type_mapper->getTokenTypeForEntityType($value->getEntityTypeId(), TRUE);
          if ($the_real_token_type !== $type) {
            unset($static_token_data[$type]);
            $static_token_data[$the_real_token_type] = $value;
          }
        }
      }
      $token_data += $static_token_data;
    }
    // Merge with global context providing global entities.
    /** @var \Drupal\Core\Plugin\Context\ContextRepositoryInterface $context_repository */
    $context_repository = \Drupal::service('context.repository');
    if (!isset(static::$contextIds)) {
      static::$contextIds = array_keys($context_repository->getAvailableContexts());
    }
    if (!empty(static::$contextIds)) {
      foreach ($context_repository->getRuntimeContexts(static::$contextIds) as $context) {
        $value = $context->getContextValue();
        if (!($value instanceof EntityInterface)) {
          // We only care for entities here.
          continue;
        }
        $type = isset($entity_token_type_mapper) ? $entity_token_type_mapper->getTokenTypeForEntityType($value->getEntityTypeId()) : $value->getEntityTypeId();
        if (!isset($token_data[$type])) {
          $token_data[$type] = $value;
        }
      }
    }

    $template_tokens = $token_service->scan($template_content);
    // Also include Mustache variables as token candidates.
    preg_match_all('/\{\{[^a-zA-Z\d\{]*([a-zA-Z\d\.]+)[^a-zA-Z\d\.\}]*\}\}/x', $template_content, $matches);
    if (!empty($matches[1])) {
      foreach ($matches[1] as $match) {
        $parts = explode('.', $match);
        $type = array_shift($parts);
        $template_tokens[$type][implode(':', $parts)] = '[' . str_replace('.', ':', $match) . ']';
      }
    }
    if (empty($template_tokens)) {
      return;
    }
    unset($matches);

    $bubbleable_metadata = new BubbleableMetadata();

    $replacements = [];
    foreach ($template_tokens as $type => $tokens) {
      $replacements += $token_service->generate($type, $tokens, $token_data, $token_options, $bubbleable_metadata);
      if (!isset($token_options['clear']) || !empty($token_options['clear'])) {
        $replacements += array_fill_keys($tokens, '');
      }
    }

    // No explicit escaping of token values is being done here, because
    // wrapped Mustache variables will be escaped by default, if not using
    // the triple {{{}}} parenthesis.
    // Optionally alter the list of replacement values.
    if (!empty($token_options['callback'])) {
      $function = $token_options['callback'];
      $function($replacements, $token_data, $token_options, $bubbleable_metadata);
    }

    $bubbleable_metadata->applyTo($element);

    // Transform tokens into Mustache variables and corresponding data.
    $mustache_variables = [];
    $mustache_data = [];
    foreach ($replacements as $token => $value) {
      $key = preg_replace('/[\s\-]+/', '_', str_replace(':', '.', substr($token, 1, -1)));
      $mustache_variables[$token] = $key;
      NestedArray::setValue($mustache_data, explode('.', $key), $value);
    }

    if (empty($element['#data'])) {
      $element['#data'] = $mustache_data;
    }
    elseif (is_array($element['#data'])) {
      $element['#data'] = NestedArray::mergeDeep($mustache_data, $element['#data']);
    }
    else {
      $element['#override_data'] = isset($element['#override_data']) ? NestedArray::mergeDeep($mustache_data, $element['#override_data']) : $mustache_data;
    }

    // Take care that standalone tokens are being wrapped by Mustache brackets,
    // and replace all token versions by their Mustache variable counterparts.
    preg_match_all('/(\{\{)?[^a-zA-Z\d\{]*(\[[^\s\[\]:]+:[^\[\]]+\])[^a-zA-Z\d\.\}]*(\}\})?/x', $template_content, $matches);
    if (!empty($matches[2])) {
      foreach ($matches[2] as $i => $token) {
        if (!isset($mustache_variables[$token])) {
          continue;
        }
        $search = $matches[0][$i];
        $mustache_variable = $mustache_variables[$token];
        if ($matches[1][$i] !== '{{') {
          $mustache_variable = '{{' . $mustache_variable;
        }
        if ($matches[3][$i] !== '}}') {
          $mustache_variable .= '}}';
        }
        $replace = str_replace($token, $mustache_variable, $search);
        $template_content = str_replace($search, $replace, $template_content);
      }
    }
  }

  /**
   * Helper function to rebuild the url with the given query parameters.
   *
   * @param \Drupal\Core\Url $url
   *   The url to rebuild.
   * @param array $new_query_params
   *   Unescaped query params to use for rebuilding.
   *
   * @return \Drupal\Core\Url
   *   The rebuilt url object.
   */
  protected static function rebuildUrl(Url $url, array $new_query_params) {
    if (!($query_params = $url->getOption('query'))) {
      $query_params = [];
    }
    $query_params = array_merge($query_params, $new_query_params);
    if ($query_params != $url->getOption('query')) {
      $url = clone $url;
      $url->setOption('query', $query_params);
    }
    return $url;
  }

}
